/**
 * @class APIBuilder.API
 * An API provides a way for a client to access your service, execute custom logic and
 * internally access the service's models and APIs, then return data back to the client.
 * Use an API to perform a custom operation on model data that is not exposed as one of the
 * standardized methods.
 *
 * ### Create an API Endpoint
 *
 * To create an API, define a new API instance with the API constructor,
 * then add the instance to the server instance and bind the API instance to the
 * API Builder service's app instance.
 *
 * Pass an implementation object, API Builder configuration object and API Builder instance to the
 * constructor. Set any API properties on the implementation object (except the ones marked
 * non-creation), but the following properties are required to be set: {@link #action}, {@link
 * #description}, {@link #group} and {@link #path}.
 *
 * After creating the API instance, add the API instance to the API Builder instance by calling the
 * API Builder instance's {@link APIBuilder#addAPI addAPI()} method, then call the API instance's
 * [bind()](#bind) method and pass it an app instance.
 *
 *      var APIBuilder = require('@axway/api-builder-runtime');
 *      var server = new APIBuilder();
 *      // since the API relies on a model,
 *      // need to wait for the model to be loaded before creating the API
 *      server.on('starting', function() {
 *          var fooAPI = new APIBuilder.API({
 *              group: 'foo',
 *              path: '/api/foo/',
 *              method: 'GET',
 *              description: 'Get all the Foo objects',
 *              model: 'foo',
 *              action: function(request, response, next) {
 *                  request.model.findAll(function(err, results) {
 *                      if (err) {
 *                          next(err);
 *                      } else {
 *                          next(null, results);
 *                      }
 *                  });
 *              }
 *          }, server.config, server);
 *          server.add(fooAPI);
 *          fooAPI.bind(server.app);
 *      }
 *      server.start();
 *
 * Alternatively, you can create an API definition file that is automatically loaded by the API
 * Builder service. For details, see the [API Builder APIs
 * guide](https://docs.axway.com/bundle/api-builder/page/docs/developer_guide/apis/index.html).
 *
 * ### Invoke an API Endpoint
 *
 * To invoke an API endpoint within the service, use the API Builder instance's
 * [getAPI()](APIBuilder#getAPI) method. An API is identified by its path and method (HTTP verb), or
 * its nickname if you defined one when creating the API. After retrieving the API instance, call
 * its [execute()](#execute) method.
 *
 *     server.getAPI('foo', 'GET').execute({}, function (err, results) {
 * 	       if (err) {
 * 	           server.logger.error(err);
 *         } else {
 * 	           // do something with results
 *         }
 *     });
 */

/**
 * @constructor
 * Creates a new API instance.
 * @param {Dictionary<APIBuilder.API>} impl Implementation object.
 * Set any API properties on the object except the ones marked non-creation. The following
 * properties must be set: {@link #action}, {@link #description}, {@link #group} and {@link #path}.
 * @param {Object} config API Builder configuration object.
 * @param {APIBuilder} apibuilder API Builder instance.
 */

/**
 * @property {Function} action
 * Logic to execute when the endpoint is invoked by the client.
 * The function is passed a request object, response object and next() function.
 * You should always make sure that the action function calls the next() function
 * regardless if the result is a success or an error.
 */
/**
 * @property {String/Array<String>} after
 * One or more Blocks to be executed after the request. Blocks are referenced by their
 * name property. If you want to execute multiple blocks, you should specify them as an
 * array of block names. If multiple blocks are specified, they are executed in the order specified.
 * Specify either `after` or `post` but not both.
 */
/**
 * @property {Object} app
 * @nonCreation
 * Express app instance.
 */
/**
 * @property {String/Array<String>} before
 * One or more Blocks to be executed before the request. Blocks are referenced by their
 * name property. If you want to execute multiple blocks, you should specify them as an
 * array of block names. If multiple blocks are specified, they are executed in the order specified.
 * Specify either `before` or `pre` but not both.
 */
/**
 * @property {Object} config
 * @nonCreation
 * Configuration object used to initialize the API instance.
 */
/**
 * @property {Function} constructor
 * Custom constructor to execute custom initialization logic. The new API instance is passed to
 * `this` in the function.
 */
/**
 * @property {String} description
 * Human-readable description of the API, which is used by the generated API documentation.
 */
/**
 * @property {String} filename
 * @nonCreation
 * File used to load the API instance.
 */
/**
 * @property {String} group
 * Logical name for the API grouping.
 */
/**
 * @property {String} method
 * HTTP verb: `GET`, `PUT`, `POST` or `DELETE`.
 */
/*
 * @property {Function} middleware
 * @nonCreation
 * Middleware block that provides access to the API's action implementation.
 */
/**
 * @property {APIBuilder.Model} model
 * The model to use for the request. Passed as the `model` property in the `request` and `response`
 * objects of the {@link #action} function.
 */
/**
 * @property {String} nickname
 * Alternative identifier for the API. By default, the API is identified by its path and HTTP verb.
 * If you do not specify a nickname, this property will be set to the path.
 */
/**
 * @property {Object} parameters
 * Input parameters required to execute the API endpoint. This is an object of key-value pairs,
 * where the key is the name of the parameter and the value is an object with the following
 * properties:
 *
 * * optional (Boolean): Determines if the parameter is optional (`true`) or required (`false`).
 * * type (String): the type of input parameter: `path` or `query`.
 * * description (String): used for generating API documentation.
 */
/**
 * @property {String} path
 * API endpoint.
 */
/**
 * @property {String} plural
 * A string used as the property name when your API endpoint returns an array.
 * By default, the plural value is the plural of the model name.
 * For example, if your model is named car, the default plural would be cars.
 */
/**
 * @property post
 * @alias #after
 */
/**
 * @property pre
 * @alias #before
 */
/**
 * @property {APIBuilder.Model} response
 * The model to use for the response. Only use the property if your request and response models are
 * different. Passed as the `modelResponse` property in the `request` and `response`
 * objects of the {@link #action} function.
 */
/**
 * @property {Object} route
 * @nonCreation
 * Express route instance that binds the endpoint to the logic.
 */
/**
 * @property {String} singular
 * A string used as the property name when your API endpoint returns a single record.
 * By default, the singular value is the name of the model.
 */
/**
 * @property {String} timestamp
 * @nonCreation
 * Datetime when the route was last synced.
 */
