/**
 * @class APIBuilder.Middleware
 * The Middleware class sets up additional features for the Express app instance, such as the adding
 * busboy, body-parser, cookie-parser and compression modules and initializing the renderer engines
 * used by the API Builder Web components. To access the middleware instance, use the API Builder
 * instance's `middleware` property.
 *
 * ### Renderer Engines
 *
 * API Builder provides renderer engines for EJS, Handlebars, HTML, Markdown and React templates.
 * The renderer engines are used by the API Builder Web component to render HTML to a web client. To
 * access the renderers, use the {@link APIBuilder.Middleware#static-method-getRendererEngine
 * getRendererEngine()} method and pass it the extension of the renderer engine to retrieve. The
 * renderer engine provides access to the module's APIs.
 *
 * To create a custom renderer engine, create an object that implements the
 * {@link APIBuilderRendererEngine#createRenderer createRenderer()} method and specifies the
 * {@link APIBuilderRendererEngine#extension extension}. Pass the object to the Middleware
 * instance's
 * {@link APIBuilder.Middleware#registerRendererEngine registerRendererEngine())} method.
 *
 * For example, to implement a renderer engine for Jade templates:
 *
 *     var jade = require('jade'),
 *         engine = {};
 *     engine.jade = jade;
 *     engine.createRenderer = function (content, filename, app) {
 *         return function(filename, opts, callback) {
 *             if (!content) {
 *                 content = require('fs').readFileSync(filename, 'utf8').toString();
 *             }
 *             callback(null, jade.render(content, opts));
 *         }
 *     };
 *     engine.extension = 'jade';
 *     server.middleware.registerRendererEngine(engine);
 *
 * Any view with a `jade` extension will be routed to the Jade renderer engine.
 */

/**
 * @class APIBuilderRendererEngine
 * A renderer engine provides access to the underlying module's APIs in addition to the ones
 * defined here. If you are implementing your own custom renderer engine, you need to implement
 * the {@link #createRenderer} method and specify the {@link #extension}.
 *
 * API Builder provides the renderer engines for the following template formats:
 *
 *  * {@link EjsRendererEngine EJS}
 *  * {@link HandlebarsRendererEngine Handlebars}
 *  * {@link HtmlRendererEngine HTML}
 *  * {@link MarkdownRendererEngine Markdown}
 *  * {@link ReactRendererEngine React}
 * @pseudo
 */
/**
 * @method createRenderer
 * Returns a function to generate the HTML. Pass the returned function the path of the template
 * file, a context object (data to pass to the template), and a callback passed an Error object (or
 * null if successful) and the HTML to render.
 * @param {String} content Template string to compile and render. Set to `null` if `skipread` is
 * @`true`.
 * @param {String} filename Path of the template file.
 * @param {Object} app Express app instance.
 * @returns {Function}
 *
 */
/**
 * @property {String} extension
 * File extension the renderer engine will handle.
 */
/**
 * @property {Boolean} skipread
 * Set to `true` to skip reading the file and converting it to a string for the `content`
 * parameter of the `createRenderer()` function.
 */
