/**
 * @class APIBuilder.Router
 * A Route is one part of the API Builder Web component, which is used to display HTML content
 * to the client. A Route defines the API endpoint used by the service and
 * the logic to execute to render the view files.
 *
 * ### Create a Router Endpoint
 *
 * To create a Router, define a new Router instance with the Router constructor,
 * then bind the Router instance to the API Builder server's app instance.
 *
 * Pass an implementation object, API Builder configuration object and API Builder instance to the
 * constructor. Set any Router properties on the implementation object except the ones marked
 * non-creation. The {@link #name} property must be set.
 *
 * To render a view file, in the `action` function, invoke the
 * [render()](http://expressjs.com/4x/api.html#res.render) method on the response object. Pass the
 * view file, located in the `./web/views` folder, to render minus the file extension and the data
 * to apply to the template. API Builder selects the renderer engine to use based on the extension
 * of the view file.
 *
 * After creating the Router instance, call its [bind()](#bind) method and pass it an app instance.
 *
 *     var foo = new APIBuilder.Router({
 *         name: 'foo',
 *         path: '/foo',
 *         method: 'GET',
 *         description: 'Render the foo view',
 *         action: function (req, res, next) {
 *             res.render('foo', {foobar: 'hello', foobaz: 'world'});
 *             next();
 *         }
 *     }, server.config, server);
 *     foo.bind(server.app);
 *
 * Alternatively, you can define your Routes using JavaScript files, which are automatically loaded
 * by the API Builder service. For details, see the [API Builder Web
 * guide](https://docs.axway.com/bundle/api-builder/page/docs/developer_guide/web/index.html).
 */

/**
 * @constructor
 * Creates a new Router instance.
 * @param {Dictionary<APIBuilder.Router>} impl Implementation object.
 * Set any Router properties on the object except the ones marked non-creation.
 * The {@link #name} property must be set.
 * @param {Object} config API Builder configuration object.
 * @param {APIBuilder} apibuilder APIBuilder instance.
 * @throws Error Missing `name` parameter.
 */

/**
 * @property {Function} action
 * Logic to execute when the endpoint is invoked by a client.
 * The function is passed a request object, response object and next() function.
 * You should always make sure that the action function calls the next() function
 * regardless if the result is a success or an error.
 */
/**
 * @property {APIBuilder} apibuilder
 * @nonCreation
 * API Builder instance associated with the Router instance.
 */
/**
 * @property {Object} config
 * @nonCreation
 * Configuration object used to initialize the Router instance.
 */
/**
 * @property {String} description
 * Human-readable description of the API, which is used by the generated API documentation.
 */
/**
 * @property {String} filename
 * @nonCreation
 * File used to load the Router instance.
 */
/**
 * @property {String} method
 * HTTP verb: `GET`, `PUT`, `POST` or `DELETE`.
 */
/**
 * @property {String} name
 * Name of the route used for identification.
 */
/**
 * @property {String} path
 * API endpoint.
 */
/**
 * @property {Object} route
 * @nonCreation
 * Express route instance that binds the endpoint to the logic.
 */
/**
 * @property {String} timestamp
 * @nonCreation
 * Datetime when the route was last synced.
 */
